
// StringJExecutor.java
// Andrew Davison, July 2007, ad@fivedots.coe.psu.ac.th

/* Compile a string as a class and execute it after
   passing a single command line argument to it.
   Uses the Compiler API and javac compiler.

   Based on an example at:
      http://www.velocityreviews.com/forums/t318697-javacompilertool.html
*/


import java.util.*;
import java.lang.reflect.*;
import javax.tools.*;
import javax.tools.JavaCompiler.CompilationTask;



public class StringJExecutor
{
  private static final String CLASS_NAME = "Foo";   // name of the generated class


  private static DiagnosticCollector<JavaFileObject> diagnostics;
       // a collector for diagnostic messages


  public static void main(String[] args)
  {
    if (args.length < 1) {
      System.out.println("No argument supplied");
      System.exit(1);
    }

    Map<String, JavaFileObject> store = 
                            new HashMap<String, JavaFileObject>();
     // maps class names to JFOs containing the classes' byte codes


    StringJFO src = makeCode(CLASS_NAME); 
    compileCode(src, store);
    runCode(CLASS_NAME, args[0], store);  // load and run
  }  // end of main()



  private static StringJFO makeCode(String className)
  /* Convert a string into a string java file object with the 
     specified class name. */
  {
    String codeStr = 
        "public class " + className + " {" +
            "public static void main(String[] args){" +
                "System.out.println(\"Hello \" + args[0]);}}";

    StringJFO src = null;
    try {
      src = new StringJFO(className + ".java", codeStr);
    }
    catch (Exception e) 
    {  System.out.println(e); 
       System.exit(1);
    }
    return src;
  } // end of makeCode()


  private static void compileCode(StringJFO src, 
                                   Map<String, JavaFileObject> store)
  // compile using input from src, output to store
  {
    CompilationTask task = makeCompilerTask(src, store);

    System.out.println("Compiling...");
    boolean hasCompiled = task.call();   // carry out the compilation
    for (Diagnostic d : diagnostics.getDiagnostics())     // list any problems
      System.out.println(d);

    if (!hasCompiled) {
      System.out.println("Compilation failed");
      System.exit(1);
    }
    else 
      System.out.println("Generated Classes: " + store.keySet());   // list generated classes
  }  // end of compileCode()



  private static CompilationTask makeCompilerTask(StringJFO src, 
                                          Map<String, JavaFileObject> store)
  /* Create a compilation operation object for compiling the 
     string java file object, src. Use a specialized java file
     manager. 
     The resulting class is saved in store under the class name. 
  */
  {
    JavaCompiler compiler = ToolProvider.getSystemJavaCompiler();
       // access the compiler provided with this JRE
    if (compiler == null) {
      System.out.println("Compiler not found");
      System.exit(1);
    }

    // create a collector for diagnostic messages
    diagnostics = new DiagnosticCollector<JavaFileObject>();

    // create a manager for the java file objects
    StandardJavaFileManager fileMan = 
            compiler.getStandardFileManager(diagnostics, null, null);

   /* forward most calls to the standard java file manager
      but put the compiled classes into the store */
    ByteJavaFileManager jfm =
          new ByteJavaFileManager<StandardJavaFileManager>(fileMan, store);

    /* create a compilation task using the supplied file manager, diagnostic
       collector, and applied to the string file object (in a list) */
    return compiler.getTask(null, jfm, diagnostics, null, null,
                                                   Arrays.asList(src));
  }  // end of makeCompilerTask()



  private static void runCode(String className, String arg,
                                          Map<String, JavaFileObject> store)
  /* Load the class, and call its main function with arg as its single command line
     argument. */
  {
    System.out.println("Running...");

    try {
      ByteClassLoader loader = new ByteClassLoader(store);   // create the class loader
      Class<?> cl = loader.loadClass(className);   // load the specified class

      // Invoke the "public static main(String[])" method
      Method mainMeth = cl.getMethod("main", new Class[] { String[].class });
      String[] methArgs = new String[] { arg };        // contains one argument
      mainMeth.invoke(null, new Object[] { methArgs }); 
    }
    catch(Exception ex)
    {  
      if (ex instanceof InvocationTargetException) {
        // ex wraps an exception thrown by the invoked method
        Throwable t = ((InvocationTargetException) ex).getTargetException();
        // System.out.println(t);
        t.printStackTrace();
      }
      else {
        // System.out.println(ex);  
        ex.printStackTrace();
      }
    }
  }  // end of runCode()


} // end of StringJExecutor class

