
// StringJViewer.java
// Andrew Davison, July 2007, ad@fivedots.coe.psu.ac.th

/* Compile a string as a class and store the byte codes in a 
   HashMap, which are then displayed using BCEL
   (http://jakarta.apache.org/bcel/).
   Uses the Compiler API and javac compiler.
*/

import java.io.*;
import java.util.*;
import javax.tools.*;
import javax.tools.JavaCompiler.CompilationTask;

import org.apache.bcel.classfile.*;   // for the BCEL API



public class StringJViewer
{
  private static final String CLASS_NAME = "Foo";   // name of the generated class

  private static DiagnosticCollector<JavaFileObject> diagnostics;
       // a collector for diagnostic messages


  public static void main(String[] args)
  {
    Map<String, JavaFileObject> store = 
                            new HashMap<String, JavaFileObject>();
     // maps class names to JFOs containing the classes' byte codes

    StringJFO src = makeCode(CLASS_NAME); 
    compileCode(src, store);
    viewStore(store); 
  }  // end of main()



  private static StringJFO makeCode(String className)
  /* Convert a string into a string java file object with the 
     specified class name. */
  {
    String codeStr = 
        "public class " + className + " {" +
            "public static void main(String[] args){" +
                "System.out.println(\"Hello \" + args[0]);}}";

    StringJFO src = null;
    try {
      src = new StringJFO(className + ".java", codeStr);
    }
    catch (Exception e) 
    {  System.out.println(e); 
       System.exit(1);
    }
    return src;
  } // end of makeCode()


  private static void compileCode(StringJFO src, 
                                   Map<String, JavaFileObject> store)
  // compile using input from src, output to store
  {
    CompilationTask task = makeCompilerTask(src, store);

    System.out.println("Compiling...");
    boolean hasCompiled = task.call();   // carry out the compilation
    for (Diagnostic d : diagnostics.getDiagnostics())     // list any problems
      System.out.println(d);

    if (!hasCompiled) {
      System.out.println("Compilation failed");
      System.exit(1);
    }
    else 
      System.out.println("Generated Classes: " + store.keySet());   // list generated classes
  }  // end of compileCode()



  private static CompilationTask makeCompilerTask(StringJFO src, 
                                          Map<String, JavaFileObject> store)
  /* Create a compilation operation object for compiling the 
     string java file object, src. Use a specialized java file
     manager. 
     The resulting class is saved in store under the class name. 
  */
  {
    JavaCompiler compiler = ToolProvider.getSystemJavaCompiler();
       // access the compiler provided with this JRE
    if (compiler == null) {
      System.out.println("Compiler not found");
      System.exit(1);
    }

    // create a collector for diagnostic messages
    diagnostics = new DiagnosticCollector<JavaFileObject>();

    // create a manager for the java file objects
    StandardJavaFileManager fileMan = 
            compiler.getStandardFileManager(diagnostics, null, null);

   /* forward most calls to the standard java file manager
      but put the compiled classes into the store */
    ByteJavaFileManager jfm =
          new ByteJavaFileManager<StandardJavaFileManager>(fileMan, store);

    /* create a compilation task using the supplied file manager, diagnostic
       collector, and applied to the string file object (in a list) */
    return compiler.getTask(null, jfm, diagnostics, null, null,
                                                   Arrays.asList(src));
  }  // end of makeCompilerTask()



  private static void viewStore(Map<String, JavaFileObject> store)
  /* Extract all the classes from the store, and display their info 
     using BCEL. */
  {
    for (String className : store.keySet()) {
      System.out.println(className + " Info");
      JavaFileObject jfo = store.get(className);      // get the java file object
      if (jfo == null)
        System.out.println("Class not found in store");
      else {
        byte[] bytes = ((ByteArrayJFO)jfo).getByteArray();    // extract the byte codes array
        viewClass(className, bytes);
      }
    }
  }  // end of viewStore()


  private static void viewClass(String className, byte[] bytes)
  {
    ClassParser p = new ClassParser(new ByteArrayInputStream(bytes), className);
    JavaClass jc = null;
    try {
      jc = p.parse();
    }
    catch(IOException e)
    {  System.out.println(e);
       return;
    }

    System.out.println(jc);   // print class structure
    // viewMethods(jc);
  }  // end of viewClass()


  private static void viewMethods(JavaClass jc)
  // display method details and their instructions
  {
    Method[] methods = jc.getMethods();
    for(int i=0; i < methods.length; i++) {
      System.out.println(methods[i]);  // print method declaration
      Code code = methods[i].getCode();
      if (code != null)
        System.out.println(code);
           // code mnemonic with symbolic references resolved
    }
  }  // end of viewMethods()


} // end of StringJViewer class

